//   Copyright (C) 2004 Scott Cartier
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//
// File packer for the NUON Boot Loader application.
//

//
// TODO: o Verify input files are correct format (TGAs especially)
//       o Strip out unnecessary data from TGA (developer section, ID, etc.)
//         before adding to the data file.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MAX_USER_FONTS 4

#define NUM_IMG_GROUPS 9

char *versionString	= "v1.3";

// File names the boot loader expects
char *filenameAppList	= "applist.txt";
char *filenameImgA	= "img1.tga";
char *filenameImgB	= "img1a.tga";
char *filenameDesc	= "app.txt";
char *filenameIntro	= "intro.tga";
char *filenameExe1	= "cd_app.cof";
char *filenameExe2	= "NUON.CD";
char *filenameBG	= "bg.tga";
char *filenameOut	= "bootload.dat";
char *filenameLog	= "log.txt";
char *filenameWAVMove	= "move.wav";
char *filenameWAVSelect	= "select.wav";

// Some globals
char **aAppNames = NULL;	// array of app directory names
int *aAppFlags   = NULL;	// array of app flags
int **aAppImgFound = NULL;      // keeps track of what app images were found
int *aFileOffset = NULL;	// array of file offsets
int *aFileSize   = NULL;	// array of file sizes
int numApps;			// number of apps in the data file
int headerSize;			// size of header (all files except executables)
int totalFileSize;		// total size of the data file
long checksumNonExe;		// accumulated checksum (all files except executables)
long *aChecksumExe = NULL;	// array of checksums for executable files
int numFiles = 0;		// number of files in the data file
int flags;			// global flags

char *strParamUserFont = "UserFont";
char *aUserFontNames[MAX_USER_FONTS];

FILE *fpAppList  = NULL;	// some file pointers
FILE *fpOutFile  = NULL;
FILE *fpDataFile = NULL;
FILE *fpLogFile  = NULL;

// ===================================================================

// Some flags for the apps
enum {
  kFlagHasImg           = (1<<0),
  kFlagHasDesc		= (1<<2),
  kFlagHasIntro		= (1<<3),
  kFlagHasExe1		= (1<<4),
  kFlagHasExe2		= (1<<5),
  kFlagHasExe		= kFlagHasExe1 | kFlagHasExe2,
  kFlagHasBG		= (1<<6),
  kFlagHasWAVMove		= (1<<7),
  kFlagHasWAVSelect	= (1<<8),
};

// ===================================================================

// Function prototypes
void getAppList         ();
void findAppFiles       ();
void createHeader       ();
void getChecksums       ();
void addLongToChecksum  (long *checksum, long val);
void getFilenameChecksum(long *checksum, char *dirname, char *filename);
void getFileChecksum    (long *checksum, char *dirname, char *filename);
int  getFileSize        (char *dirname, char *filename);
void writeOutputFile    ();
int  writeLong          (FILE *fp, long val);
int  writeFilename      (FILE *fp, char *dirname, char *filename);
int  writeFile          (FILE *fp, char *dirname, char *filename);
void verifyFile         ();
long getChecksum        (unsigned char *buffer, int bytes);
void quit               ();

// ===================================================================

// I like clean main functions :)
int main()
{
  printf("\nNUON Boot Packer %s (built %s)", versionString, __DATE__);
  
  getAppList();
  findAppFiles();
  createHeader();
  getChecksums();
  writeOutputFile();
  verifyFile();
  
  quit();
  
  return 0;
}

//===================================================================
// getAppList()
//
// Parse the applist.txt file to get the app directories and font
// names.
//===================================================================

void getAppList()
{
  int i, j;
  int curApp;
  char line[80];
  int len;
  
  fpAppList = fopen(filenameAppList, "r");
  
  if (!fpAppList) {
    printf("\n-E- Could not find %s", filenameAppList);
    quit();
  }
  
  // Clear the font name list
  for (i = 0; i < MAX_USER_FONTS; i++) {
    aUserFontNames[i] = NULL;
  }
  
  printf("\nReading %s", filenameAppList);
  
  // Parse the app list to get the number of apps.  Also read the
  // font names along the way.
  numApps = 0;
  while (!feof(fpAppList)) {
    line[0] = 0;
    fgets(line, 80, fpAppList);
    while (isspace(line[strlen(line)-1])) {
      line[strlen(line)-1] = 0;
    }
    
    switch (line[0]) {
    case '#':  break;  // comment - do nothing
    case '+':
      // Check for user font names
      if ((strncmp(&line[1], strParamUserFont, strlen(strParamUserFont)) == 0) &&
	  (line[strlen(strParamUserFont)+1] >= '1') &&
	  (line[strlen(strParamUserFont)+1] < ('1' + MAX_USER_FONTS)) &&
	  (line[strlen(strParamUserFont)+2] == '=')) {
	i = line[strlen(strParamUserFont)+1] - '1';
	if (aUserFontNames[i]) {
	  printf("\n-E- Multiple fonts for user font #%d", i + 1);
	  quit();
	}
	aUserFontNames[i] =
	  (char*) malloc(sizeof(char) * (strlen(&line[strlen(strParamUserFont)+3]) + 1));
	if (!aUserFontNames[i]) {
	  printf("\n-E- Could not allocate %d bytes for user font name %s",
		 strlen(&line[strlen(strParamUserFont)+3]) + 1, &line[strlen(strParamUserFont)+3]);
	  quit();
	}
	strcpy(aUserFontNames[i], &line[strlen(strParamUserFont)+3]);
      }
      break;
    case '/':  numApps++;  break;
    }
  }
  
  // Allocate space for the given number of apps
  len = sizeof(char*) * numApps;
  aAppNames = (char **) malloc (len);
  
  if (!aAppNames) {
    printf("\n-E- Could not allocate %d bytes for app name array", len);
    quit();
  }
  
  for (i = 0; i < numApps; i++) aAppNames[i] = NULL;
  
  // Allocate space for the app flags
  len = sizeof(int) * numApps;
  aAppFlags = (int *) malloc (len);
  
  if (!aAppFlags) {
    printf("\n-E- Could not allocate %d bytes for app flag array", len);
    quit();
  }
  
  for (i = 0; i < numApps; i++) aAppFlags[i] = 0;
  
  // Re-parse the file, filling in array of app names
  fseek(fpAppList, 0, SEEK_SET);
  curApp = 0;
  while (!feof(fpAppList)) {
    line[0] = 0;
    fgets(line, 80, fpAppList);
    if (line[0]) {
      if (line[strlen(line) - 1] < ' ') line[strlen(line) - 1] = 0;
      
      switch (line[0]) {
      case '#':  break;  // comment - do nothing
      case '+':  break;  // parameter - do nothing
      case '/':
	len = sizeof(char) * (strlen(line) + 1);
	aAppNames[curApp] = (char *) malloc (len);
	if (!aAppNames[curApp]) {
	  printf("\n-E- Could not allocate %d bytes for app name %d", len, curApp);
	  quit();
	}
	strncpy(aAppNames[curApp], line, strlen(line));
	aAppNames[curApp][strlen(line)] = 0;
	curApp++;
	break;
      }
    }
  }
  
  // Allocate space for the img present flags
  len = sizeof(int*) * numApps;
  aAppImgFound = (int **) malloc (len);
  
  if (!aAppImgFound) {
    printf("\n-E- Could not allocate %d bytes for app img found array", len);
    quit();
  }
  
  for (i = 0; i < numApps; i++) {
    len = sizeof(int) * NUM_IMG_GROUPS;
    aAppImgFound[i] = (int*) malloc (len);
    if (!aAppImgFound[i]) {
      printf("\n-E- Could not allocate %d bytes for app img found array (%d)", len, i);
      quit();
    }
    for (j = 0; j < NUM_IMG_GROUPS; j++) {
      aAppImgFound[i][j] = 0;
    }
  }

  fclose(fpAppList);
  fpAppList = NULL;
}

//===================================================================
// findAppFiles()
//
// Look for the various optional files and set flags if found.
//===================================================================

void findAppFiles()
{
  int i;
  int imgGroupNum, imgNum;
  int curApp;
  char filename[80];
  
  printf("\n");
  printf("\nFinding files");
  printf("\n   %s", filenameAppList);
  
  flags = 0;
  
  // Check for Background
  sprintf(filename, "%s", filenameBG);
  fpDataFile = fopen(filename, "rb");
  if (fpDataFile) flags |= kFlagHasBG;
  else {
    filename[strlen(filename) - 3] = 'j';
    filename[strlen(filename) - 2] = 'p';
    filename[strlen(filename) - 1] = 'g';
    fpDataFile = fopen(filename, "rb");
    if (fpDataFile) flags |= kFlagHasBG;
  }
  if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
  if (flags & kFlagHasBG) printf("\n   %s", filename);
  
  // Check for sounds
  sprintf(filename, "%s", filenameWAVMove);
  fpDataFile = fopen(filename, "rb");
  if (fpDataFile) flags |= kFlagHasWAVMove;
  if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
  if (flags & kFlagHasWAVMove) printf("\n   %s", filename);
  
  sprintf(filename, "%s", filenameWAVSelect);
  fpDataFile = fopen(filename, "rb");
  if (fpDataFile) flags |= kFlagHasWAVSelect;
  if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
  if (flags & kFlagHasWAVSelect) printf("\n   %s", filename);
  
  // Check for user fonts
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      sprintf(filename, "%s", aUserFontNames[i]);
      fpDataFile = fopen(filename, "rb");
      if (!fpDataFile) {
	printf("\n-E- Could not find font %s", filename);
	quit();
      } else {
	fclose(fpDataFile);
	fpDataFile = NULL;
      }
      printf("\n   %s", filename);
    }
  }
  
  
  // Cycle through the apps
  for (curApp = 0; curApp < numApps; curApp++) {
    
    printf("\n   App: %s", &aAppNames[curApp][1]);
    // Check for Exe
    sprintf(filename, "%s/%s", &aAppNames[curApp][1], filenameExe1);
    fpDataFile = fopen(filename, "rb");
    if (!fpDataFile) {
      sprintf(filename, "%s/%s", &aAppNames[curApp][1], filenameExe2);
      fpDataFile = fopen(filename, "rb");
      
      if (fpDataFile) {
	aAppFlags[curApp] |= kFlagHasExe2;
      }
    } else {
      aAppFlags[curApp] |= kFlagHasExe1;
    }
    
    if (fpDataFile) {
      fclose(fpDataFile);
      fpDataFile = NULL;
    }
    
    if (aAppFlags[curApp] & kFlagHasExe) printf("\n      %s", filename);
    else printf("\n-W- Didn't find %s or %s - excluding app", filenameExe1, filenameExe2);
    
    // Don't bother checking for the other files if the executable doesn't exist
    if (aAppFlags[curApp] & kFlagHasExe) {
      
      // Check for images in up to 9 groups
      for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
	
	// First check the original filename "img#.tga" and "img#.jpg"
	filenameImgA[3] = '1' + imgGroupNum;
	sprintf(filename, "%s/%s", &aAppNames[curApp][1], filenameImgA);
	fpDataFile = fopen(filename, "rb");
	if (fpDataFile) {
	  aAppImgFound[curApp][imgGroupNum] |= 1<<0;
	  aAppFlags[curApp] |= kFlagHasImg;
	} else {
	  filename[strlen(filename) - 3] = 'j';
	  filename[strlen(filename) - 2] = 'p';
	  filename[strlen(filename) - 1] = 'g';
	  fpDataFile = fopen(filename, "rb");
	  if (fpDataFile) {
	    aAppImgFound[curApp][imgGroupNum] |= 1<<0;
	    aAppFlags[curApp] |= kFlagHasImg;
	  }
	}
	if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
	if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) printf("\n      %s", filename);
	
	// Now check for the new filenames "img#a.tga" and "img#a.jpg"
	filenameImgB[3] = '1' + imgGroupNum;
	for (imgNum = 0; imgNum < 26; imgNum++) {
	  filenameImgB[4] = 'a' + imgNum;
	  sprintf(filename, "%s/%s", &aAppNames[curApp][1], filenameImgB);
	  fpDataFile = fopen(filename, "rb");
	  if (fpDataFile) {
	    aAppImgFound[curApp][imgGroupNum] |= 1<<(imgNum+1);
	    aAppFlags[curApp] |= kFlagHasImg;
	  } else {
	    filename[strlen(filename) - 3] = 'j';
	    filename[strlen(filename) - 2] = 'p';
	    filename[strlen(filename) - 1] = 'g';
	    fpDataFile = fopen(filename, "rb");
	    if (fpDataFile) {
	      aAppImgFound[curApp][imgGroupNum] |= 1<<(imgNum+1);
	      aAppFlags[curApp] |= kFlagHasImg;
	    }
	  }
	  if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
	  if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) printf("\n      %s", filename);
	}
      }
      
      // Check for description
      sprintf(filename, "%s/%s", &aAppNames[curApp][1], filenameDesc);
      fpDataFile = fopen(filename, "rb");
      if (fpDataFile) aAppFlags[curApp] |= kFlagHasDesc;
      if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
      if (aAppFlags[curApp] & kFlagHasDesc) printf("\n      %s", filename);
      
      // Check for intro image
      sprintf(filename, "%s/%s", &aAppNames[curApp][1], filenameIntro);
      fpDataFile = fopen(filename, "rb");
      if (fpDataFile) aAppFlags[curApp] |= kFlagHasIntro;
      else {
	filename[strlen(filename) - 3] = 'j';
	filename[strlen(filename) - 2] = 'p';
	filename[strlen(filename) - 1] = 'g';
	fpDataFile = fopen(filename, "rb");
	if (fpDataFile) aAppFlags[curApp] |= kFlagHasIntro;
      }
      if (fpDataFile) { fclose(fpDataFile);  fpDataFile = NULL; }
      if (aAppFlags[curApp] & kFlagHasIntro) printf("\n      %s", filename);
    }
  }
}

//===================================================================
// createHeader()
//
// The first section computes the size of the header.  The header is
// defined as everything that gets loaded upon startup.  That 
// includes all the app images & description text.  It does not
// include the executables since they are too large to all fit at
// once.
//
// The second section computes all the file sizes and offsets.
//===================================================================

void createHeader()
{
  int i;
  int curApp;
  int len;
  int fileNum;
  int numExes = 0;
  int error;
  int imgGroupNum, imgNum;
  
  //
  // Compute header size.
  //
  
  headerSize = 0;
  
  // Get the size of the header
  headerSize += 8;		// "BOOTLOAD" signature
  headerSize += 4;		// Total file size
  headerSize += 4;		// Size of header in bytes
  headerSize += 4;		// Header checksum
  headerSize += 4;		// Number of files
  
  // Start with the (required) applist.txt
  numFiles++;
  headerSize += 4;		// File offset
  headerSize += 4;		// File size
  headerSize += (strlen(filenameAppList) + 1 + 0x3) & ~0x3;
  
  // Increment if we have a background image
  if (flags & kFlagHasBG) {
    numFiles++;
    headerSize += 4;		// File offset
    headerSize += 4;		// File size
    headerSize += (strlen(filenameBG) + 1 + 0x3) & ~0x3;
  }
  
  // Increment if we have sounds
  if (flags & kFlagHasWAVMove) {
    numFiles++;
    headerSize += 4;		// File offset
    headerSize += 4;		// File size
    headerSize += (strlen(filenameWAVMove) + 1 + 0x3) & ~0x3;
  }
  if (flags & kFlagHasWAVSelect) {
    numFiles++;
    headerSize += 4;		// File offset
    headerSize += 4;		// File size
    headerSize += (strlen(filenameWAVSelect) + 1 + 0x3) & ~0x3;
  }
  
  // Increment if we have fonts
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      numFiles++;
      headerSize += 4;		// File offset
      headerSize += 4;		// File size
      headerSize += (strlen(aUserFontNames[i]) + 1 + 0x3) & ~0x3;
    }
  }
  
  for (curApp = 0; curApp < numApps; curApp++) {
    for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
      if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) {
	numFiles++;
	headerSize += 4;		// File offset
	headerSize += 4;		// File size
	headerSize += (strlen(aAppNames[curApp]) + strlen(filenameImgA) + 1 + 0x3) & ~0x3;
      }
      for (imgNum = 0; imgNum < 26; imgNum++) {
	if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) {
	  numFiles++;
	  headerSize += 4;		// File offset
	  headerSize += 4;		// File size
	  headerSize += (strlen(aAppNames[curApp]) + strlen(filenameImgB) + 1 + 0x3) & ~0x3;
	}
      }
    }
    if (aAppFlags[curApp] & kFlagHasDesc) {
      numFiles++;
      headerSize += 4;		// File offset
      headerSize += 4;		// File size
      headerSize += (strlen(aAppNames[curApp]) + strlen(filenameDesc) + 1 + 0x3) & ~0x3;
    }
    if (aAppFlags[curApp] & kFlagHasIntro) {
      numFiles++;
      headerSize += 4;		// File offset
      headerSize += 4;		// File size
      headerSize += (strlen(aAppNames[curApp]) + strlen(filenameIntro) + 1 + 0x3) & ~0x3;
    }
  }
  
  error = 0;
  for (curApp = 0; curApp < numApps; curApp++) {
    if (aAppFlags[curApp] & kFlagHasExe) {
      numFiles++;
      numExes++;
      headerSize += 4;		// File offset
      headerSize += 4;		// File size
      if (aAppFlags[curApp] & kFlagHasExe1) {
	headerSize += (strlen(aAppNames[curApp]) + strlen(filenameExe1) + 1 + 0x3) & ~0x3;
      } else {
	headerSize += (strlen(aAppNames[curApp]) + strlen(filenameExe2) + 1 + 0x3) & ~0x3;
      }
    } else {
      printf("\n-E- App %s does not have an executable", aAppNames[curApp]);
      error = 1;
    }
  }
  
  if (error) quit();
  
  printf("\n");
  printf("\nNum files found: %d", numFiles);
  
  // Allocate enough room for all the executable checksums
  len = sizeof(long) * numExes;
  aChecksumExe = (long *) malloc (len);
  if (!aChecksumExe) {
    printf("\n-E- Could not allocate %d bytes for EXE checksum array", len);
    quit();
  }
  
  // Allocate space for the file sizes & offsets.
  // Add room for one extra due to the way offsets are computed.
  len = sizeof(int) * (numFiles + 1);
  aFileOffset = (int *) malloc (len);
  if (!aFileOffset) {
    printf("\n-E- Could not allocate %d bytes for file offset array", len);
    quit();
  }
  len = sizeof(int) * numFiles;
  aFileSize = (int *) malloc (len);
  if (!aFileSize) {
    printf("\n-E- Could not allocate %d bytes for file size array", len);
    quit();
  }
  
  //
  // Compute the file offsets.
  //
  
  printf("\n");
  printf("\nFile offsets");
  
  // applist.txt
  fileNum = 0;
  aFileOffset[fileNum] = headerSize;
  aFileSize[fileNum] = getFileSize(NULL, filenameAppList);
  headerSize += aFileSize[fileNum];
  aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
  printf("\n   %s = %d", filenameAppList, aFileOffset[fileNum]);
  fileNum++;
  
  if (flags & kFlagHasBG) {
    // bg.jpg (or tga)
    aFileSize[fileNum] = getFileSize(NULL, filenameBG);
    headerSize += aFileSize[fileNum];
    aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
    printf("\n   %s = %d", filenameBG, aFileOffset[fileNum]);
    fileNum++;
  }
  if (flags & kFlagHasWAVMove) {
    // move.wav
    aFileSize[fileNum] = getFileSize(NULL, filenameWAVMove);
    headerSize += aFileSize[fileNum];
    aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
    printf("\n   %s = %d", filenameWAVMove, aFileOffset[fileNum]);
    fileNum++;
  }
  if (flags & kFlagHasWAVSelect) {
    // select.wav
    aFileSize[fileNum] = getFileSize(NULL, filenameWAVSelect);
    headerSize += aFileSize[fileNum];
    aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
    printf("\n   %s = %d", filenameWAVSelect, aFileOffset[fileNum]);
    fileNum++;
  }
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      aFileSize[fileNum] = getFileSize(NULL, aUserFontNames[i]);
      headerSize += aFileSize[fileNum];
      aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
      printf("\n   %s = %d", aUserFontNames[i], aFileOffset[fileNum]);
      fileNum++;
    }
  }
  
  for (curApp = 0; curApp < numApps; curApp++) {
    if (aAppFlags[curApp] & (kFlagHasImg | kFlagHasDesc | kFlagHasIntro)) {
      printf("\n   App: %s", &aAppNames[curApp][1]);
    }
    
    for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
      if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) {
	filenameImgA[3] = '1' + imgGroupNum;
	aFileSize[fileNum] = getFileSize(aAppNames[curApp], filenameImgA);
	headerSize += aFileSize[fileNum];
	aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
	printf("\n      %s = %d", filenameImgA, aFileOffset[fileNum]);
	fileNum++;
      }
      for (imgNum = 0; imgNum < 26; imgNum++) {
	if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) {
	  filenameImgB[3] = '1' + imgGroupNum;
	  filenameImgB[4] = 'a' + imgNum;
	  aFileSize[fileNum] = getFileSize(aAppNames[curApp], filenameImgB);
	  headerSize += aFileSize[fileNum];
	  aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
	  printf("\n      %s = %d", filenameImgB, aFileOffset[fileNum]);
	  fileNum++;
	}
      }
    }
    if (aAppFlags[curApp] & kFlagHasDesc) {
      // app.txt
      aFileSize[fileNum] = getFileSize(aAppNames[curApp], filenameDesc);
      headerSize += aFileSize[fileNum];
      aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
      printf("\n      %s = %d", filenameDesc, aFileOffset[fileNum]);
      fileNum++;
    }
    if (aAppFlags[curApp] & kFlagHasIntro) {
      // intro.jpg (or tga)
      aFileSize[fileNum] = getFileSize(aAppNames[curApp], filenameIntro);
      headerSize += aFileSize[fileNum];
      aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum];
      printf("\n      %s = %d", filenameIntro, aFileOffset[fileNum]);
      fileNum++;
    }
  }
  
  // Sanity check that the final offset is equal to the header size computed earlier
  if (headerSize != aFileOffset[fileNum]) {
    printf("\n-E- INTERNAL ERROR - Header size != last offset (%d != %d)",
	   headerSize, aFileOffset[fileNum]);
    quit();
  }
  
  // Now compute the size and offset of the executables
  for (curApp = 0; curApp < numApps; curApp++) {
    printf("\n   App: %s", &aAppNames[curApp][1]);
    if (aAppFlags[curApp] & kFlagHasExe1) {
      aFileSize[fileNum] = getFileSize(aAppNames[curApp], filenameExe1);
      aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum] + 4; // add 4 for checksum
      printf("\n      %s = %d", filenameExe1, aFileOffset[fileNum]);
      fileNum++;
    } else if (aAppFlags[curApp] & kFlagHasExe2) {
      aFileSize[fileNum] = getFileSize(aAppNames[curApp], filenameExe2);
      aFileOffset[fileNum+1] = aFileOffset[fileNum] + aFileSize[fileNum] + 4; // add 4 for checksum
      printf("\n      %s = %d", filenameExe2, aFileOffset[fileNum]);
      fileNum++;
    }
  }
  
  totalFileSize = aFileOffset[fileNum];
  
  if (fileNum != numFiles) {
    printf("\n-E- INTERNAL ERROR - fileNum != numFiles (%d != %d)", fileNum, numFiles);
    quit();
  }
  
}

//===================================================================
// getChecksums()
//
// Computes the checksums used to verify the file at run time.  There
// is one checksum associated with the header (everything but the
// executables).  There is also a checksum associated with each
// executable.
//===================================================================

void getChecksums()
{
  int i;
  long buf = 0;
  int curApp;
  int curFile = 0;
  int exeFile = 0;
  int imgGroupNum, imgNum;
  
  fpLogFile = fopen(filenameLog, "w");
  
  checksumNonExe = 0;
  
  // Signature
  buf = ('B' << 24) | ('O' << 16) | ('O' << 8) | ('T' << 0);
  addLongToChecksum(&checksumNonExe, buf);
  
  buf = ('L' << 24) | ('O' << 16) | ('A' << 8) | ('D' << 0);
  addLongToChecksum(&checksumNonExe, buf);
  
  // File info
  addLongToChecksum(&checksumNonExe, totalFileSize);
  addLongToChecksum(&checksumNonExe, headerSize);
  addLongToChecksum(&checksumNonExe, numFiles);
  
  // 
  // Create checksums for the file names, offsets, and sizes
  // 
  
  // applist.txt
  getFilenameChecksum(&checksumNonExe, NULL, filenameAppList);
  addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
  addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
  curFile++;
  
  if (flags & kFlagHasBG) {
    // bg.jpg
    getFilenameChecksum(&checksumNonExe, NULL, filenameBG);
    addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
    addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
    curFile++;
  }
  if (flags & kFlagHasWAVMove) {
    // move.wav
    getFilenameChecksum(&checksumNonExe, NULL, filenameWAVMove);
    addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
    addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
    curFile++;
  }
  if (flags & kFlagHasWAVSelect) {
    // select.wav
    getFilenameChecksum(&checksumNonExe, NULL, filenameWAVSelect);
    addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
    addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
    curFile++;
  }
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      getFilenameChecksum(&checksumNonExe, NULL, aUserFontNames[i]);
      addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
      addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
      curFile++;
    }
  }
  
  for (curApp = 0; curApp < numApps; curApp++) {
    for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
      if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) {
 	filenameImgA[3] = '1' + imgGroupNum;
	getFilenameChecksum(&checksumNonExe, aAppNames[curApp], filenameImgA);
	addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
	addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
	curFile++;
      }

      for (imgNum = 0; imgNum < 26; imgNum++) {
	if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) {
	  filenameImgB[3] = '1' + imgGroupNum;
	  filenameImgB[4] = 'a' + imgNum;
	  getFilenameChecksum(&checksumNonExe, aAppNames[curApp], filenameImgB);
	  addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
	  addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
	  curFile++;
	}
      }
    }
    if (aAppFlags[curApp] & kFlagHasDesc) {
      // app.txt
      getFilenameChecksum(&checksumNonExe, aAppNames[curApp], filenameDesc);
      addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
      addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
      curFile++;
    }
    if (aAppFlags[curApp] & kFlagHasIntro) {
      // intro.jpg
      getFilenameChecksum(&checksumNonExe, aAppNames[curApp], filenameIntro);
      addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
      addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
      curFile++;
    }
  }
  
  for (curApp = 0; curApp < numApps; curApp++) {
    // cd_app.cof
    if (aAppFlags[curApp] & kFlagHasExe1) {
      getFilenameChecksum(&checksumNonExe, aAppNames[curApp], filenameExe1);
    }
    if (aAppFlags[curApp] & kFlagHasExe2) {
      getFilenameChecksum(&checksumNonExe, aAppNames[curApp], filenameExe2);
    }
    addLongToChecksum(&checksumNonExe, aFileOffset[curFile]);
    addLongToChecksum(&checksumNonExe, aFileSize[curFile]);
    curFile++;
  }
  
  //
  // Now add to the checksum for the files themselves.
  //
  
  getFileChecksum(&checksumNonExe, NULL, filenameAppList);
  if (flags & kFlagHasBG       ) getFileChecksum(&checksumNonExe, NULL, filenameBG       );
  if (flags & kFlagHasWAVMove  ) getFileChecksum(&checksumNonExe, NULL, filenameWAVMove  );
  if (flags & kFlagHasWAVSelect) getFileChecksum(&checksumNonExe, NULL, filenameWAVSelect);
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      getFileChecksum(&checksumNonExe, NULL, aUserFontNames[i]);
    }
  }
  for (curApp = 0; curApp < numApps; curApp++) {
    if (aAppFlags[curApp] & kFlagHasExe) {
      for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
	if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) {
	  filenameImgA[3] = '1' + imgGroupNum;
	  getFileChecksum(&checksumNonExe, &aAppNames[curApp][1], filenameImgA);
	}
	for (imgNum = 0; imgNum < 26; imgNum++) {
	  if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) {
	    filenameImgB[3] = '1' + imgGroupNum;
	    filenameImgB[4] = 'a' + imgNum;
	    getFileChecksum(&checksumNonExe, &aAppNames[curApp][1], filenameImgB);
	  }
	}
      }
      if (aAppFlags[curApp] & kFlagHasDesc) {
	getFileChecksum(&checksumNonExe, &aAppNames[curApp][1], filenameDesc);
      }
      if (aAppFlags[curApp] & kFlagHasIntro) {
	getFileChecksum(&checksumNonExe, &aAppNames[curApp][1], filenameIntro);
      }
    }
  }
  
  if (fpLogFile) fclose(fpLogFile);
  fpLogFile = NULL;
  
  //
  // Finally, get the checksums for the exe files
  //
  
  for (curApp = 0; curApp < numApps; curApp++) {
    aChecksumExe[exeFile] = 0;
    if (aAppFlags[curApp] & kFlagHasExe1) {
      getFileChecksum(&aChecksumExe[exeFile++], &aAppNames[curApp][1], filenameExe1);
    }
    if (aAppFlags[curApp] & kFlagHasExe2) {
      getFileChecksum(&aChecksumExe[exeFile++], &aAppNames[curApp][1], filenameExe2);
    }
  }
  
  if (curFile != numFiles) {
    printf("\n-E- INTERNAL ERROR: curFile (%d) != numFiles (%d)", curFile, numFiles);
    quit();
  }
  if (exeFile != numApps) {
    printf("\n-E- INTERNAL ERROR: exeFile (%d) != numApps (%d)", exeFile, numApps);
    quit();
  }
  
  if (fpLogFile) fclose(fpLogFile);
  fpLogFile = NULL;
}

//===================================================================
// addLongToChecksum()
//
// Adds a long value to the current checksum.
//===================================================================
void addLongToChecksum(long *checksum, long val)
{
  *checksum += val;
  if (fpLogFile) fprintf(fpLogFile, "\n+%08X = %08X", val, *checksum);
}


//===================================================================
// getFilenameChecksum()
//
// Creates a checksum for the given file name.
//===================================================================
void getFilenameChecksum(long *checksum, char *dirname, char *filename)
{
  char fullPath[80];
  unsigned char *ptr;
  int numChars;
  long val;
  FILE *fp;
  
  memset(fullPath, 0, 80);
  if (dirname) {  strcat(fullPath, &dirname[1]);  strcat(fullPath, "/");  }
  strcat(fullPath, filename);
  
  // For images, must find either a .jpg or .tga
  fp = fopen(fullPath, "r");
  if (!fp && (strncmp(&fullPath[strlen(fullPath) - 3], "tga", 3) == 0)) {
    fullPath[strlen(fullPath) - 3] = 'j';
    fullPath[strlen(fullPath) - 2] = 'p';
    fullPath[strlen(fullPath) - 1] = 'g';
    fp = fopen(fullPath, "r");
  }
  if (!fp) {
    printf("\n-E- INTERNAL ERROR: Could not find img %s/%s", dirname, filename);
    quit();
  }
  
  if (fpLogFile) fprintf(fpLogFile, "\n\nFILENAME: %s", fullPath);
  
  ptr = (unsigned char *) fullPath;
  for (numChars = strlen(fullPath); numChars > 0; numChars -= 4) {
    val = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3] << 0);
    *checksum += val;
    ptr += 4;
    if (fpLogFile) fprintf(fpLogFile, "\n+%08X = %08X", val, *checksum);
  }
  
  if (fpLogFile) fprintf(fpLogFile, "\n");
}


//===================================================================
// getFileChecksum()
//
// Creates a checksum for the given file.
//===================================================================
void getFileChecksum(long *checksum, char *dirname, char *filename)
{
  char fullPath[80];
  unsigned char mylong[4];
  long val;
  
  memset(fullPath, 0, 80);
  if (dirname) {  strcat(fullPath, dirname);  strcat(fullPath, "/");  }
  strcat(fullPath, filename);
  
  if (fpLogFile) fprintf(fpLogFile, "\n\nFILE: %s", fullPath);
  
  fpDataFile = fopen(fullPath, "rb");
  if (!fpDataFile) {
    fullPath[strlen(fullPath) - 3] = 'j';
    fullPath[strlen(fullPath) - 2] = 'p';
    fullPath[strlen(fullPath) - 1] = 'g';
    fpDataFile = fopen(fullPath, "rb");
    if (!fpDataFile) {
      printf("\n-E- Could not open output file %s for reading", fullPath);
      quit();
    }
  }
  
  while (!feof(fpDataFile)) {
    memset(mylong, 0, 4);
    if (fread(mylong, 1, 4, fpDataFile) != 0) {
      val = (mylong[0] << 24) | (mylong[1] << 16) | (mylong[2] << 8) | (mylong[3] << 0);
      *checksum += val;
      if (fpLogFile) fprintf(fpLogFile, "\n+%08X = %08X", val, *checksum);
    }
  }
  
  fclose(fpDataFile);
  fpDataFile = NULL;
}


//===================================================================
// getFileSize()
//
// Returns the size of a file.
//===================================================================
int getFileSize(char *dirname, char *filename)
{
  int size;
  char fullPath[80];
  char buf[32];
  FILE *fp;
  
  fullPath[0] = 0;
  if (dirname) {
    strcpy(fullPath, &dirname[1]);
    strcat(fullPath, "/");
  }
  strcat(fullPath, filename);
  
  // For images, must find either a .jpg or .tga
  fpDataFile = fopen(fullPath, "rb");
  if (!fpDataFile) {
    if (strncmp(&fullPath[strlen(fullPath) - 3], "tga", 3) == 0) {
      fullPath[strlen(fullPath) - 3] = 'j';
      fullPath[strlen(fullPath) - 2] = 'p';
      fullPath[strlen(fullPath) - 1] = 'g';
      fpDataFile = fopen(fullPath, "rb");
    }
    if (!fpDataFile) {
      printf("\n-E- Could not open %s in %s", fullPath, __func__);
      quit();
    }
  }
  
  fseek(fpDataFile, 0, SEEK_END);
  size = ftell(fpDataFile);
  fclose(fpDataFile);
  fpDataFile = NULL;
  
  // Account for NULL at end of text files
  if (strncmp(&fullPath[strlen(fullPath) - 3], "txt", 3) == 0) {
    size++;
  }

  // Round up to dword boundary
  size = (size + 0x3) & ~0x3;

  return size;
}


//===================================================================
// writeOutputFile()
//
// Writes the final output data file.
//===================================================================
void writeOutputFile()
{
  int i;
  char buf[80];
  int curApp;
  int fileNum = 0;
  int exeNum = 0;
  int headerWritten = 0;
  int totalWritten;
  int imgGroupNum, imgNum;
  
  // Open the output file for real and start writing
  fpOutFile = fopen(filenameOut, "wb");
  
  if (!fpOutFile) {
    printf("\n-E- Could not open output file %s for writing", filenameOut);
    quit();
  }
  
  // Signature
  strcpy(buf, "BOOTLOAD");
  headerWritten += fwrite(buf, 1, 8, fpOutFile);
  
  // File info
  headerWritten += writeLong(fpOutFile, totalFileSize );
  headerWritten += writeLong(fpOutFile, headerSize    );
  headerWritten += writeLong(fpOutFile, -checksumNonExe);
  headerWritten += writeLong(fpOutFile, numFiles      );
  
  // App list
  headerWritten += writeFilename(fpOutFile, NULL, filenameAppList);
  headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
  headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
  fileNum++;
  
  // Background
  if (flags & kFlagHasBG) {
    headerWritten += writeFilename(fpOutFile, NULL, filenameBG);
    headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
    headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
    fileNum++;
  }
  
  // Sounds
  if (flags & kFlagHasWAVMove) {
    headerWritten += writeFilename(fpOutFile, NULL, filenameWAVMove);
    headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
    headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
    fileNum++;
  }
  if (flags & kFlagHasWAVSelect) {
    headerWritten += writeFilename(fpOutFile, NULL, filenameWAVSelect);
    headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
    headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
    fileNum++;
  }
  
  // Fonts
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      headerWritten += writeFilename(fpOutFile, NULL, aUserFontNames[i]);
      headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
      headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
      fileNum++;
    }
  }
  
  // App filenames (non-exe)
  for (curApp = 0; curApp < numApps; curApp++) {
    for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
      if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) {
	filenameImgA[3] = '1' + imgGroupNum;
	headerWritten += writeFilename(fpOutFile, aAppNames[curApp], filenameImgA);
	headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
	headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
	fileNum++;
      }
      for (imgNum = 0; imgNum < 26; imgNum++) {
	if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) {
	  filenameImgB[3] = '1' + imgGroupNum;
	  filenameImgB[4] = 'a' + imgNum;
	  headerWritten += writeFilename(fpOutFile, aAppNames[curApp], filenameImgB);
	  headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
	  headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
	  fileNum++;
	}
      }
    }
    if (aAppFlags[curApp] & kFlagHasDesc) {
      headerWritten += writeFilename(fpOutFile, aAppNames[curApp], filenameDesc);
      headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
      headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
      fileNum++;
    }
    if (aAppFlags[curApp] & kFlagHasIntro) {
      headerWritten += writeFilename(fpOutFile, aAppNames[curApp], filenameIntro);
      headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
      headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
      fileNum++;
    }
  }
  
  // App filenames (exe)
  for (curApp = 0; curApp < numApps; curApp++) {
    if (aAppFlags[curApp] & kFlagHasExe1) {
      headerWritten += writeFilename(fpOutFile, aAppNames[curApp], filenameExe1);
    }
    if (aAppFlags[curApp] & kFlagHasExe2) {
      headerWritten += writeFilename(fpOutFile, aAppNames[curApp], filenameExe2);
    }
    headerWritten += writeLong(fpOutFile, aFileOffset[fileNum]);
    headerWritten += writeLong(fpOutFile, aFileSize[fileNum]);
    fileNum++;
  }
  
  // Sanity check
  if (fileNum != numFiles) {
    printf("\n-E- INTERNAL ERROR - fileNum != numFiles (%d != %d)", fileNum, numFiles);
    quit();
  }
  
  //
  // Now write the files themselves.
  //
  
  // applist.txt
  headerWritten += writeFile(fpOutFile, NULL, filenameAppList);
  
  // bg.jpg
  if (flags & kFlagHasBG) headerWritten += writeFile(fpOutFile, NULL, filenameBG);
  
  // move.wav
  if (flags & kFlagHasWAVMove) headerWritten += writeFile(fpOutFile, NULL, filenameWAVMove);
  
  // select.wav
  if (flags & kFlagHasWAVSelect) headerWritten += writeFile(fpOutFile, NULL, filenameWAVSelect);
  
  // User fonts
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) {
      headerWritten += writeFile(fpOutFile, NULL, aUserFontNames[i]);
    }
  }
  
  // App images/description
  for (curApp = 0; curApp < numApps; curApp++) {
    for (imgGroupNum = 0; imgGroupNum < NUM_IMG_GROUPS; imgGroupNum++) {
      if (aAppImgFound[curApp][imgGroupNum] & (1<<0)) {
	filenameImgA[3] = '1' + imgGroupNum;
	headerWritten += writeFile(fpOutFile, &aAppNames[curApp][1], filenameImgA);
      }
      for (imgNum = 0; imgNum < 26; imgNum++) {
	if (aAppImgFound[curApp][imgGroupNum] & (1<<(imgNum+1))) {
	  filenameImgB[3] = '1' + imgGroupNum;
	  filenameImgB[4] = 'a' + imgNum;
	  headerWritten += writeFile(fpOutFile, &aAppNames[curApp][1], filenameImgB);
	}
      }
    }
    if (aAppFlags[curApp] & kFlagHasDesc) {
      headerWritten += writeFile(fpOutFile, &aAppNames[curApp][1], filenameDesc);
    }
    if (aAppFlags[curApp] & kFlagHasIntro) {
      headerWritten += writeFile(fpOutFile, &aAppNames[curApp][1], filenameIntro);
    }
  }
  
  // Sanity check
  if (headerWritten != headerSize) {
    printf("\n-E- INTERNAL ERROR - Header written != header size (%d != %d)",
	   headerWritten, headerSize);
    quit();
  }
  
  totalWritten = headerWritten;
  
  // Now write the executable and its checksum
  for (curApp = 0; curApp < numApps; curApp++) {
    totalWritten += writeLong(fpOutFile, -aChecksumExe[exeNum++]);
    if (aAppFlags[curApp] & kFlagHasExe1) {
      totalWritten += writeFile(fpOutFile, &aAppNames[curApp][1], filenameExe1);
    }
    if (aAppFlags[curApp] & kFlagHasExe2) {
      totalWritten += writeFile(fpOutFile, &aAppNames[curApp][1], filenameExe2);
    }
  }
  
  if (totalWritten != totalFileSize) {
    printf("\n-E- INTERNAL ERROR - Total written != total file size (%d != %d)",
	   totalWritten, totalFileSize);
    quit();
  }
  
  if (fpOutFile) fclose(fpOutFile);
  fpOutFile = NULL;
}


//===================================================================
// writeLong()
//
// Write a long to the file and return the number of bytes written.
//===================================================================
int writeLong(FILE *fp, long val)
{
  char buf[4];
  
  buf[0] = (val >> 24) & 0xFF;
  buf[1] = (val >> 16) & 0xFF;
  buf[2] = (val >>  8) & 0xFF;
  buf[3] = (val >>  0) & 0xFF;
  
  return fwrite(buf, 1, 4, fp);
}


//===================================================================
// writeFilename()
//
// Write a file name to the file and return the number of bytes 
// written.
//===================================================================
int writeFilename(FILE *fp, char *dirname, char *filename)
{
  char fullPath[80];
  int numChars;
  
  fullPath[0] = 0;
  if (dirname) {
    strcat(fullPath, &dirname[1]);
    strcat(fullPath, "/");
  }
  strcat(fullPath, filename);
  
  // For images, must find either a .jpg or .tga
  fpDataFile = fopen(fullPath, "rb");
  if (!fpDataFile) {
    if (strncmp(&fullPath[strlen(fullPath) - 3], "tga", 3) == 0) {
      fullPath[strlen(fullPath) - 3] = 'j';
      fullPath[strlen(fullPath) - 2] = 'p';
      fullPath[strlen(fullPath) - 1] = 'g';
      fpDataFile = fopen(fullPath, "rb");
    }
    if (!fpDataFile) {
      printf("\n-E- Could not open %s in getFileSize()", fullPath);
      quit();
    }
  }
  fclose(fpDataFile);
  fpDataFile = NULL;
  
  numChars = fwrite(fullPath, 1, strlen(fullPath), fpOutFile);
  // Round up to multiple of 4 bytes
  fullPath[0] = 0;
  numChars += fwrite(fullPath, 1, 1, fpOutFile);
  while (numChars & 0x3) numChars += fwrite(fullPath, 1, 1, fpOutFile);
  
  return numChars;
}


//===================================================================
// writeFile()
//
// Write a file to the output file and return the number of bytes 
// written.
//===================================================================
int writeFile(FILE *fp, char *dirname, char *filename)
{
  char fullPath[80];
  int fileSize;
  unsigned char *buf;
  int numChars;
  
  fullPath[0] = 0;
  if (dirname) {
    strcpy(fullPath, dirname);
    strcat(fullPath, "/");
  }
  strcat(fullPath, filename);
  
  // For images, must find either a .jpg or .tga
  fpDataFile = fopen(fullPath, "rb");
  if (!fpDataFile) {
    if (strncmp(&fullPath[strlen(fullPath) - 3], "tga", 3) == 0) {
      fullPath[strlen(fullPath) - 3] = 'j';
      fullPath[strlen(fullPath) - 2] = 'p';
      fullPath[strlen(fullPath) - 1] = 'g';
      fpDataFile = fopen(fullPath, "rb");
    }
    if (!fpDataFile) {
      printf("\n-E- Could not open %s in %s", fullPath, __func__);
      quit();
    }
  }
  
  // Get the size of the file
  fseek(fpDataFile, 0, SEEK_END);
  fileSize = ftell(fpDataFile);
  rewind(fpDataFile);
  
  // Allocate enough room to read in the whole file (it shouldn't be that large)
  buf = (unsigned char *) malloc (sizeof(unsigned char) * fileSize);
  if (!buf) {
    printf("\n-E- Could not allocate %d bytes for file buffer", fileSize);
    quit();
  }
  
  // Read the input file
  numChars = fread(buf, 1, fileSize, fpDataFile);
  if (numChars != fileSize) {
    printf("\n-E- Didn't read the expected number of chars (file size=%d, chars read=%d)",
	   fileSize, numChars);
    free(buf);
    quit();
  }
  
  // Write the output file
  numChars = fwrite(buf, 1, fileSize, fp);
  if (numChars != fileSize) {
    printf("\n-E- Didn't write the expected number of chars (file size=%d, chars written=%d)",
	   fileSize, numChars);
    free(buf);
    quit();
  }
  
  // For text files, append a null
  buf[0] = 0;
  if (strncmp(&fullPath[strlen(fullPath) - 3], "txt", 3) == 0) 
    numChars += fwrite(buf, 1, 1, fpOutFile);
  
  // Round up to a multiple of 4 bytes
  while (numChars & 0x3)
    numChars += fwrite(buf, 1, 1, fpOutFile);
  
  free(buf);
  fclose(fpDataFile);
  fpDataFile = NULL;
  
  return numChars;
}


//===================================================================
// verifyFile()
//
// Sanity check the output file to make sure everything is okay for
// the boot loader.
//===================================================================
void verifyFile()
{
  unsigned char buf[16];
  long bufLong;
  int actualFileSize;
  int bytesRead;
  int totalFileSize, headerSize;
  long compare;
  long actualChecksum;
  int loc;
  unsigned char *bufDataFile;
  
  
  fpLogFile = fopen(filenameLog, "a");
  
  if (fpLogFile) fprintf(fpLogFile, "\n\n*****\n");
  
  fpDataFile = fopen(filenameOut, "rb");
  if (!fpDataFile) {
    printf("\n-E- 0");
    quit();
  }
  
  bytesRead = fread(buf, 1, 16, fpDataFile);
  
  if (bytesRead != 16) {
    printf("\n-E- 1");
    quit();
    
  } else {
    // Read the first 16 bytes
    
    // Now check BOOT
    compare = ('B' << 24) | ('O' << 16) | ('O' << 8) | ('T' << 0);
    bufLong = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | (buf[3] << 0);
    if (bufLong != compare) {
      printf("\n-E- 2");
      quit();
      
    } else {
      // BOOT matched
      // Now check LOAD
      
      compare = ('L' << 24) | ('O' << 16) | ('A' << 8) | ('D' << 0);
      bufLong = (buf[4] << 24) | (buf[5] << 16) | (buf[6] << 8) | (buf[7] << 0);
      if (bufLong != compare) {
	printf("\n-E- 3");
	quit();
	
      } else {
	// BOOT matched
	// LOAD matched
	// Now check the file size
	
	totalFileSize = (buf[8] << 24) | (buf[9] << 16) | (buf[10] << 8) | (buf[11] << 0);
	headerSize = (buf[12] << 24) | (buf[13] << 16) | (buf[14] << 8) | (buf[15] << 0);
	
	loc = ftell(fpDataFile);
	fseek(fpDataFile, 0, SEEK_END);
	actualFileSize = ftell(fpDataFile);
	fseek(fpDataFile, loc, SEEK_SET);
	if (actualFileSize != totalFileSize) {
	  printf("\n-E- 4");
	  quit();
	} else {
	  // BOOT matched
	  // LOAD matched
	  // File size matched
	  // Now read the full header
	  
	  // Allocate space for the full header, leaving room to pad to a dword boundary
	  bufDataFile = (unsigned char *) malloc(sizeof(char) * (headerSize + 3));
	  
	  rewind(fpDataFile);
	  bytesRead = fread(bufDataFile, 1, headerSize, fpDataFile);
	  
	  if (bytesRead != headerSize) {
	    printf("\n-E- 5");
	    quit();
	    
	  } else {
	    // BOOT matched
	    // LOAD matched
	    // File size matched
	    // Successfully read full header
	    // Now check checksum
	    
	    // pad zeros to a dword boundary
	    while (bytesRead & 0x3) bufDataFile[bytesRead++] = 0;
	    
	    //	      headerChecksum = *((long *) &bufDataFile[16]);
	    actualChecksum = getChecksum(bufDataFile, bytesRead);
	    
	    if (actualChecksum) {
	      printf("\n-E- Checksum error during verification");
	      quit();
	      
	    } else {
	      // BOOT matched
	      // LOAD matched
	      // File size matched
	      // Successfully read full header
	      // Checksum matched
	      
	      printf("\nVERIFIED!");
	      
	    }  // if (headerChecksum != actualChecksum)
	  }  // if (bytesRead != headerSize)
	}  // if (actualFileSize != totalFileSize)
      }  // if (bufLong != compare)
    }  // if (bufLong != compare)
  }  // if (bytesRead != 16)
  
  if (fpOutFile) fclose(fpOutFile);
  fpOutFile = NULL;
  
  free(bufDataFile);
  
  if (fpLogFile) fclose(fpLogFile);
  fpLogFile = NULL;
}


//===================================================================
// getChecksum()
//
// Compute the checksum over the given memory range.
//===================================================================
long getChecksum(unsigned char *buffer, int bytes)
{
  unsigned char *ptr;
  long checksum = 0;
  long val;
  int wordnum = 0;
  long filechecksum = 0;
  
  bytes = (bytes + 3) & ~0x3;  // round up to dword boundary, if necessary
  
  for (ptr = buffer; bytes; bytes -= 4, ptr += 4, wordnum++) {
    //    printf("%08X  ", *ptr);
    val = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3] << 0);
    checksum += val;
    if (wordnum == 4) filechecksum = val;
    if (fpLogFile) {
      fprintf(fpLogFile, "\n+%08X = %08X", val, checksum-filechecksum);
      if (wordnum == 4) { filechecksum = val; fprintf(fpLogFile, "  CHECKSUM"); }
    }
  }
  
  return checksum;
}


//===================================================================
// quit()
//
// Generic exit routine that cleans up memory and file pointers.
//===================================================================
void quit()
{
  int i;
  
  if (aAppNames) {
    for (i = 0; i < numApps; i++)
      if (aAppNames[i]) free(aAppNames[i]);
    free(aAppNames);
  }
  
  if (aAppFlags  ) free(aAppFlags  );
  if (aFileOffset) free(aFileOffset);
  if (aFileSize  ) free(aFileSize  );
  
  if (fpAppList ) fclose(fpAppList );
  if (fpDataFile) fclose(fpDataFile);
  if (fpOutFile ) fclose(fpOutFile );
  if (fpLogFile ) fclose(fpLogFile );
  
  for (i = 0; i < MAX_USER_FONTS; i++) {
    if (aUserFontNames[i]) free(aUserFontNames[i]);
  }
  
  printf("\n");
  exit(0);
}
